from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from httpx import AsyncClient
import aiofiles
from pathlib import Path
from random import choice
from urllib.parse import urljoin
from tqdm import tqdm
from asyncio import get_running_loop, as_completed, gather, get_event_loop

from APP.utils.profiles import select_profile
from APP.models.post import Post
from APP.utils.make_async import make_async, async_as_completed
from APP.config import headers, timeout, base_urls
from APP.utils.sanitize_path import sanitize_name


async def download(profile=None):
    if profile is None:
        profile = select_profile()
    loop = get_running_loop()
    engine = create_engine(profile.db_string)
    Session = sessionmaker(bind=engine)
    try:
        async with AsyncClient() as client:
            download_tasks = []
            with Session() as session:
                results = session.query(Post).filter(Post.downloaded == False).all()
                for result in results:
                    download_tasks.append(loop.create_task(final_download(result.post_id, profile, client)))

                pbar = tqdm(total=len(download_tasks), desc="Downloading", unit="files")
                async for task in async_as_completed(download_tasks):
                    pbar.update(1)
                pbar.close()
    except Exception as e:
        print(e)
    finally:
        engine.dispose()

async def final_download(result, profile, client):
    e = create_engine(profile.db_string)
    S = sessionmaker(bind=e)
    with S() as s:
        result = s.query(Post).filter(Post.post_id == result).first()
        result.load()
        async for item in make_async(result.media):
            save_location = Path(await sanitize_name(str(profile.save_location)), await sanitize_name(result.category), await sanitize_name(result.subject), await sanitize_name(result.message), await sanitize_name(result.author))
            save_location.mkdir(parents=True, exist_ok=True)
            tried_urls = []
            file_name = item.split('/')[-1]
            file_path = Path(save_location, file_name)
            while True:
                try:
                    possible_urls = [u for u in base_urls if u not in tried_urls]
                    used_url = choice(possible_urls)
                    tried_urls.append(used_url)
                    url = urljoin(used_url, item)
                    async with client.stream('GET', url, headers=headers, timeout=timeout) as r:
                        async with aiofiles.open(file_path, 'wb') as f:
                            async for chunk in r.aiter_bytes():
                                await f.write(chunk)

                    result.downloaded = True
                    result.dump()
                    s.commit()
                    break
                except Exception as e:
                    while len(tried_urls) < len(base_urls):
                        pass
                    else:
                        break



from asyncio import Semaphore
from tqdm import tqdm



from APP.utils.make_async import make_async
from APP.scrapers.find_boards import find_boards
from APP.scrapers.get_board_links import get_board_link
from APP.utils.profiles import select_profile
from APP.scrapers.index_posts import index_posts




async def scrape(profile=None):
    if profile is None:
        profile = select_profile()
    # Define a semaphore to limit the number of concurrent requests
    requests_semaphore = Semaphore(profile.max_concurrent_requests)

    async for category in make_async(profile.categories):
        boards = await find_boards(category, requests_semaphore)
        async for board in make_async(boards):
            print(f"Gathering links from {category}.", end="\r")
            board_sets = [await get_board_link(board) for board in boards]
            async for set in make_async(board_sets):
                async for title, link in make_async(set):
                    await index_posts(link, category, title, requests_semaphore, profile)


async def daemon_mode():
    print("This isn't available yet.")
    pass



import ssl
from httpx import AsyncClient
from bs4 import BeautifulSoup
from random import choice
from urllib.parse import urljoin


from APP.config import base_urls
from APP.config import headers, timeout

# Pick a random base url to use to avoid hitting the same url over and over. (All urls hit the same server but cloudflare
# will block you if you hit the same url too many times in a row.) Adding proxies soon.
base_url = choice(base_urls)


async def find_boards(board:str, requests_semaphore):
    async with requests_semaphore:
        async with AsyncClient() as client:
            tried_urls = []
            r = True
            while r:
                try:
                    possible_urls = [url for url in base_urls if url not in tried_urls]
                    used_url = choice(possible_urls)
                    tried_urls.append(used_url)
                    board_url = urljoin(used_url, board)
                    response = await client.get(urljoin(base_url, board), headers=headers, timeout=timeout, follow_redirects=True)
                    soup = BeautifulSoup(response.content, 'html.parser')
                    boards = soup.find_all(class_="opCell")
                    return boards
                except ssl.SSLError:
                    pass

                except Exception as e:
                    r = False
                    raise e



# The base urls of the site.
base_urls:list['str'] = [
    'https://anonib.al',
    'https://anonib.pk',
]

# The headers used for requests.
headers: dict[str:str] = {
    "authority": "anonib.al",
    "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "accept-language": "en-US,en;q=0.9",
    "cache-control": "max-age=0",
    "cookie": "captchaid=653ef276f7546e9342c1dbc20GZBeDuDBgmYPF26w5RZXqo/3zDe9r6POCx8KTIPnKuO4cwBbPdl03EZUJxlEsLf0N4kFJrw9JvXqM+UIutijw==; captchaexpiration=Mon, 30 Oct 2023 00:02:58 GMT",
    "if-modified-since": "Sun, 29 Oct 2023 23:15:01 GMT",
    "sec-ch-ua": '^\^"Google Chrome^\^";v=^\^"117^\^", ^\^"Not;A=Brand^\^";v=^\^"8^\^", ^\^"Chromium^\^";v=^\^"117^\^"',
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-platform": '^\^"Windows^\^"',
    "sec-fetch-dest": "document",
    "sec-fetch-mode": "navigate",
    "sec-fetch-site": "none",
    "sec-fetch-user": "?1",
    "upgrade-insecure-requests": "1",
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
}

# The timeout used for requests.
timeout: int = 120

# Folder to save profiles in.
profiles_folder: str = 'profiles'

# Letter subs is used to correct messages. The key is replaced with the value.
letter_subs: dict[str:str] = {
    "0":"o",
    "7":"t",
    "4":"h",
    "1":"l",
    "!":"i",
    "3":"e",
    "@":"a",
    "<":"",
    ">":"",
    "$":"s",
}

chars_replace: dict[str:str] = {

    "?":"",
    "$":"",
    "#":"",
    "*":"",
    "&":"",
    "(":"",
    ")":"",

}

# Set the max message length to use in names and folder names.
max_message_len: int = 200

# All categories

all_categories: list[str] = [

    'ak',
    'al',
    'ar',
    'az',
    'ca',
    'co',
    'ct',
    'dc',
    'de',
    'fl',
    'ga',
    'hi',
    'ia',
    'id',
    'il',
    'in',
    'ks',
    'ky',
    'la',
    'ma',
    'md',
    'me',
    'mi',
    'mn',
    'mo',
    'ms',
    'mt',
    'nc',
    'nd',
    'ne',
    'nh',
    'nj',
    'nm',
    'nv',
    'ny',
    'oh',
    'ok',
    'or',
    'pa',
    'ri',
    'sc',
    'sd',
    'tn',
    'tx',
    'ut',
    'va',
    'vt',
    'wa',
    'wi',
    'wv',
    'wy',

    # Other Cats

    "a",
    "an",
    "ass",
    "azn",
    "b",
    "bdsm",
    "bj",
    "bt",
    "c",
    "cam",
    "cb",
    "cf",
    "ci",
    "coe",
    "cosp",
    "cut",
    "cw",
    "eb",
    "et",
    "es",
    "fa",
    "fe",
    "gf",
    "gif",
    "gin",
    "hc",
    "he",
    "hy",
    "les",
    "lt",
    "mg",
    "mil",
    "milf",
    "mod",
    "mu",
    "pan",
    "paycam",
    "pi",
    "pl",
    "pr",
    "red",
    "soc",
    "tat",
    "tblr",
    "tr",
    "tt",
    "v",
    "wc",
    "ygwbt",
    "yt",
    "au",
    "can",
    "ger",
    "uk",
    "rmw",
    "btc",
    "pol"

]




from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy import create_engine, Column, Integer, Text, Boolean
from json import loads, dumps
from pathlib import Path


Base = declarative_base()


class Post(Base):

    __tablename__ = 'posts'

    id = Column(Integer, primary_key=True, autoincrement=True, nullable=False, unique=True)
    post_id = Column(Text, nullable=False, unique=True)
    message = Column(Text, nullable=False)
    author = Column(Text, nullable=False)
    category = Column(Text, nullable=False)
    subject = Column(Text, nullable=False)
    media = Column(Text, nullable=True)
    downloaded = Column(Boolean, nullable=False, default=False)

    def __init__(self, post_id, message, author, category, subject, media):
        self.post_id = post_id
        self.message = message
        self.author = author
        self.category = category
        self.subject = subject
        self.media = media
        self.downloaded = False

    def save(self, profile):
        self.save_location = str(profile.save_location)
        self.media = dumps(self.media)
        Base.metadata.create_all(create_engine(profile.db_string))
        engine = create_engine(profile.db_string)
        Session = sessionmaker(bind=engine)
        with Session() as session:
            results = session.query(Post).filter(Post.post_id == self.post_id).first()
            if results is None:
                print(f"Saving post: {self.post_id}")
                session.add(self)
                session.commit()
            else:
                print(f"Post: {self.post_id} already exists in database.")

    def __repr__(self):
        identity = (f"Post: {self.post_id}\n"
                    f"Message: {self.message}\n"
                    f"Author: {self.author}\n"
                    f"Category: {self.category}\n"
                    f"Subject: {self.subject}\n"
                    f"Media: {self.media}\n"
                    f"Downloaded: {self.downloaded}")
        return identity


    def load(self):
        self.media = loads(self.media)


    def dump(self):
        self.media = dumps(self.media)



import webbrowser


from APP.utils.profiles import select_profile, create_profile
from APP.scrapers.scrape import scrape, daemon_mode
from APP.utils.download import download


async def scrape_and_download():
    profile = select_profile()
    await scrape(profile)
    await download(profile)


async def edit_profile():
    profile = select_profile()
    profile.edit()

async def open_github_site():
    try:
        webbrowser.open('https://github.com/taux1c/AIBS')
    except:
        print('Failed to open github site but you can find it here: https://github.com/taux1c/AIBS')

async def adjust_request_rate():
    profile = select_profile()
    profile.set_request_rate()
async def close_app():
    quit()
async def main_menu():
    try:
        webbrowser.open('https://www.buymeacoffee.com/taux1c')
    except:
        pass
    options = {
        '1': create_profile,
        '2': edit_profile,
        '3': scrape,
        '4': download,
        '5': scrape_and_download,
        '6': daemon_mode,
        '7': open_github_site,
        '8': adjust_request_rate,
        '9': close_app
    }
    while True:
        for key, value in options.items():
            print(f'{key}: {value.__name__}')
        choice = input('Enter your choice: ')
        if choice not in options.keys():
            print('Invalid choice.\nPlease try again.')
        else:
            break
    if choice is not None:
        try:
            await options[choice]()
        except:
            raise





from json import dumps, loads
from pathlib import Path


from APP.config import profiles_folder


class User:

    def __init__(self):
        self.index = 0
        self.display_name = None
        self.categories = None
        self.save_location = None
        self.max_concurrent_requests = 10
        self.db_string = None


    def build(self):
        self.display_name = input("What can we call you?: ")
        self.categories = input("What categories do you want to follow? (Separate with comma.): ").split(',')
        self.save_location = Path(input("Where do you want to save your data?: "))
        self.db_string = f"sqlite:///{self.save_location}/posts.db"
        self.save()

    def save(self):
        profile_name = f"{self.display_name}.json"
        self.save_location.mkdir(parents=True, exist_ok=True)
        self.save_location = str(self.save_location)
        with open(Path(profiles_folder) / profile_name, 'w') as f:
            f.write(dumps(self.__dict__, indent=6))

    def load(self, profile_name):
        with open(Path(profiles_folder) / profile_name, 'r') as f:
            self.__dict__ = loads(f.read())
            self.save_location = Path(self.save_location)

    def __repr__(self):
        return f"display_name={self.display_name}, categories={self.categories}, save_location={self.save_location}"

    def set_max_concurrent_requests(self):
        self.max_concurrent_requests = int(input("How many concurrent requests do you want to make? (Default is 10): "))
        self.save()


def select_profile():
    if not Path(profiles_folder).exists():
        Path(profiles_folder).mkdir(parents=True, exist_ok=True)
    profiles = [profile for profile in Path(profiles_folder).iterdir() if profile.is_file() and profile.exists()]
    if len(profiles) == 0:
        print("No profiles found. Please create one.")
        quit()
    print("Select a profile:")
    for i, profile in enumerate(profiles):
        print(f"{i+1}. {profile.stem}")
    profile_selection = int(input("Enter the number of the profile you want to use: "))
    profile = profiles[profile_selection-1]
    p = User()
    p.load(profile.name)
    return p


async def create_profile():
    p = User()
    p.build()



import string
from re import sub
from APP.config import letter_subs, chars_replace, max_message_len

async def fix_message(message):
    for k,v in letter_subs.items():
        message = message.replace(k,v)
    return message

async def sanitize_name(message):
    if len(message) > max_message_len:
        message = message[-max_message_len:]
    # Using the last part of the message because this is used to sanitize file names and it will remove the extension.
    # if you use the first part; so we would need to account for that if we want to use the first part.
    message = sub(r'[,…<>:"/\\|?*\n]', '', message)
    message = message.replace("..","")
    if message.endswith('.'):
        message = message.replace('.','')
    message = message.strip()
    message = await sanitize_patch(message)
    return message

async def sanitize_patch(filename):
    # Replace invalid characters with underscores
    valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    sanitized_filename = ''.join(c if c in valid_chars else '_' for c in filename)
    return sanitized_filename



import ssl
from httpx import AsyncClient
from bs4 import BeautifulSoup as BS
from random import choice
from urllib.parse import urljoin


from APP.config import base_urls, headers, timeout
from APP.utils.make_async import make_async
from APP.models.post import Post


async def index_posts(board_url, category, subject, requests_semaphore, profile):
    async with requests_semaphore:
        async with AsyncClient() as client:
            tried_urls = []
            r = True
            while r:
                try:
                    possible_urls = [url for url in base_urls if url not in tried_urls]
                    used_url = choice(possible_urls)
                    tried_urls.append(used_url)
                    board_url = urljoin(used_url, board_url)
                    response = await client.get(board_url, headers=headers, timeout=timeout, follow_redirects=True)
                    r = False
                except ssl.SSLError:
                    print(f'SSL Error with {used_url} Retrying...')
                    pass

                except Exception as e:
                    raise e
                try:
                    soup = BS(response.content, 'lxml')
                    posts_container = soup.find(class_="opCell")
                    origin_post = posts_container.find(class_="innerOP")
                    media_replies = [p for p in posts_container.find_all(class_="postCell") if p.find_all(class_="panelUploads")]
                    media_posts = media_replies + [origin_post] if origin_post.find_all(class_="panelUploads") else media_replies
                    async for post in make_async(media_posts):
                        message = post.find(class_="divMessage").text
                        author = post.find(class_="linkName").text
                        post_id = post.find(class_="deletionCheckBox")['name']
                        post_category = category
                        post_subject = subject
                        post_media = [x['href'] for x in post.find_all(class_="nameLink")]
                        px = Post(post_id, message, author, category, subject, post_media)
                        px.save(profile)
                except Exception as e:
                    raise e




from setuptools import setup

setup(
    name='AIBS',
    version='0.0.51',
    packages=['APP','APP.utils', 'APP.scrapers', 'APP.models'],
    url='https://github.com/taux1c/AIBS',
    license='MIT',
    author='Taux1c',
    author_email='taux1c.software@protonmail.com',
    description='A simple, database-centered scraper for anonib.',
    install_requires=[
        'aiofiles~=23.2.1',
        'SQLAlchemy~=2.0.23',
        'httpx~=0.25.1',
        'tqdm~=4.66.1',
        'beautifulsoup4~=4.12.2',
    ],
    long_description="""AnonIB Web Scraper This is a database-centric web scraper designed for extracting information from AnonIB forums. The scraper is freely available for use, and contributions from the community are welcomed and appreciated.

Usage Users are free to utilize this tool for their needs without any mandatory requirement for donations. However, if you find the tool helpful, you can make voluntary contributions through the link provided when the browser opens.

Contributing Feel free to contribute to this project by submitting pull requests. Your contributions can help enhance the functionality and efficiency of the web scraper for the benefit of the community.

Disclaimer This tool is intended for educational purposes only. It is not affiliated with AnonIB in any way, and any trademarks or intellectual property rights associated with AnonIB belong to their rightful owners. Any laws or terms of service violated by the use of this tool are solely the responsibility of the user, as this tool is meant strictly for educational purposes and not intended for actual use.

The tool is provided as-is without any warranty. I, as the creator, accept no liability for the usage of this tool.

Feel free to further modify this disclaimer to ensure it aligns with your intentions and the legal requirements of your project.""",
    entry_points={
        'console_scripts': [
            'anonib=APP.main:main'
        ]
    }
)




from asyncio import as_completed
async def make_async(item):
    if isinstance(item, list):
        for i in item:
            if i is not None:
                yield i
    elif isinstance(item, dict):
        for k, v in item.items():
            if v is not None and k is not None:
                yield k, v

async def async_as_completed(tasks):
    for task in as_completed(tasks):
        if task is not None:
            await task
            yield task



from APP.config import base_urls

async def get_board_link(board):
    board_title = board.find(class_="linkName").text
    board_url = board.find(class_="linkReply")["href"]
    return {board_title:board_url}


